#include "kinematics.h"
#include "peakmap.h"

#define PRTHR_CUTREL	0.75
			
/**********************************************************************/

double Resonance(double xran, double m, double g, double Emax, double& ssq)
{
  const static double prefac = 1./(2*M_PI);
  static double rmin, rran;
  
  if(g > 0.0)
  {
    rmin = atan(-m/g);
    rran = atan((Emax*Emax - m*m)/(m*g)) - rmin;
    ssq = sqrt(m*m + m*g*tan(rran*xran + rmin));
//    cout << ":";
    return prefac * rran * (sqr(ssq*ssq-m*m) + m*m*g*g)/(m*g);
  }
  else
  {
//    cout << ".";
    ssq = m;
    return prefac;
  }
}

double jacResonance(double ps, double m, double g, double Emax)
{
  const static double prefac = 1./(2*M_PI);
  static double rran;
  
  if(g > 0.0)
  {
    rran = atan((Emax*Emax-m*m)/(m*g)) - atan(-m/g);
    return 1/(prefac * rran * (sqr(ps-m*m) + m*m*g*g)/(m*g));
  }
  else
    return 1/prefac * PseudoDelta(ps, m*m);
}


double Flat(double xran, double Emax, double& ssq)
{
  const static double prefac = 2./(2*M_PI);

  ssq = Emax * xran;
  return prefac * Emax * ssq;
}

double jacFlat(double ps, double Emax)
{
  const static double prefac = 2./(2*M_PI);
  
  return 1/(prefac * Emax * sqrt(ps));
}


double SubDec(double *xran, double ssq,
  	      double m1, vector4 k1, double m2, vector4 k2)
{
  const static double prefac = 2. * 2.*M_PI;

  return prefac *
	  _TwoDecay(ssq, m1, m2, 2*xran[0]-1, xran[1]*2*M_PI, k1, k2);
}

double jacDec(double ms, double m1s, double m2s)
{
  return (4.*M_PI)/(16*M_PI*M_PI) * sqrt(lambda(ms, m1s, m2s)) / (2*ms);
}
  

double Mapoo(double *xran, double s,
  	     double m1, vector4 k1, double m2, vector4 k2)
{
  double ssq = sqrt(s);
  return 1/(2*s) * SubDec(xran, ssq, m1, k1, m2, k2);
}

double jacoo(double s, double m1, cvector4 k1, double m2, cvector4 k2)
{
  return (2*s) / jacDec(s, m1*m1, m2*m2);
}


double Mapooo(double *xran, double s,
	double m1, vector4 k1, double m2, vector4 k2, double m3, vector4 k3)
{
  static double q1[4], r2[4], r3[4];
  double fac, s1sq, ssq = sqrt(s);
  
  fac = Flat(xran[0], ssq-m1, s1sq);
  fac *= 1/(2*s) * SubDec(xran+1, ssq, s1sq, q1, m1, k1)
	  	 * SubDec(xran+3, s1sq, m2, r2, m3, r3);
  Boost(r2, q1, k2);
  Boost(r3, q1, k3);
  return fac;
}

double jacooo(double s,
	double m1, cvector4 k1, double m2, cvector4 k2, double m3, cvector4 k3)
{
  double s1 = m2*m2 + m3*m3 + 2*Dot(k2,k3);
  return 2*s * jacFlat(s1, sqrt(s)-m1) /
	  	( jacDec(s, s1, m1*m1)
	  	* jacDec(s1, m2*m2, m3*m3) );
}


double Maprooo(double *xran, double s, double m, double g,
	double m1, vector4 k1, double m2a, vector4 k2a, double m2b, vector4 k2b)
{
  static double q1[4], r2a[4], r2b[4];
  double fac, s1sq, ssq = sqrt(s);
  
  fac = Resonance(xran[0], m, g, ssq-m1, s1sq);
  fac *= 1/(2*s) * SubDec(xran+1, ssq, s1sq, q1, m1, k1)
	  	 * SubDec(xran+3, s1sq, m2a, r2a, m2b, r2b);
  Boost(r2a, q1, k2a);
  Boost(r2b, q1, k2b);
  return fac;
}

double jacrooo(double s, double m, double g, double m1, cvector4 k1,
		double m2a, cvector4 k2a, double m2b, cvector4 k2b)
{
  double ssq = sqrt(s);
  double s1 = s + m1*m1 - 2* ssq * k1[0];	      // s1 = ((p1+p2) - k1)^2
  return 2*s * jacResonance(s1, m, g, ssq-m1) /
	  	( jacDec(s, s1, m1*m1)
	  	* jacDec(s1, m2a*m2a, m2b*m2b) );
}


double Mapoooo(double *xran, double s,
	double m1a, vector4 k1a, double m1b, vector4 k1b,
	double m2a, vector4 k2a, double m2b, vector4 k2b)
{
  static double q1[4], q2[4], r1a[4], r1b[4], r2a[4], r2b[4];
  double fac, s1sq, s2sq, ssq = sqrt(s);

  fac = Flat(xran[0], ssq, s1sq);
  fac *= Flat(xran[1], ssq-s1sq, s2sq);
  fac *= 1/(2*s) * SubDec(xran+2, ssq, s1sq, q1, s2sq, q2)
  		 * SubDec(xran+4, s1sq, m1a, r1a, m1b, r1b)
	  	 * SubDec(xran+6, s2sq, m2a, r2a, m2b, r2b);
  Boost(r1a, q1, k1a);
  Boost(r1b, q1, k1b);
  Boost(r2a, q2, k2a);
  Boost(r2b, q2, k2b);
  return fac;
}

double jacoooo(double s,
	double m1a, cvector4 k1a, double m1b, cvector4 k1b,
	double m2a, cvector4 k2a, double m2b, cvector4 k2b)
{
  double s1 = m1a*m1a + m1b*m1b + 2*Dot(k1a,k1b),
         s2 = m2a*m2a + m2b*m2b + 2*Dot(k2a,k2b),
         ssq = sqrt(s);
  return 2*s * jacFlat(s1, ssq)
	     * jacFlat(s2, ssq-sqrt(s1)) /
	  	( jacDec(s, s1, s2)
	  	* jacDec(s1, m1a*m1a, m1b*m1b)
	  	* jacDec(s2, m2a*m2a, m2b*m2b) );
}


double Maproooo(double *xran, double s, double m1, double g1,
	double m0a, vector4 k0a, double m0b, vector4 k0b,
	double m1a, vector4 k1a, double m1b, vector4 k1b)
{
  static double q1[4], q0[4], r1a[4], r1b[4], r0a[4], r0b[4];
  double fac, s1sq, s0sq, ssq = sqrt(s);

  fac = Resonance(xran[0], m1, g1, ssq, s1sq);
  fac *= Flat(xran[1], ssq-s1sq, s0sq);
  fac *= 1/(2*s) * SubDec(xran+2, ssq, s1sq, q1, s0sq, q0)
  		 * SubDec(xran+4, s1sq, m1a, r1a, m1b, r1b)
	  	 * SubDec(xran+6, s0sq, m0a, r0a, m0b, r0b);
  Boost(r1a, q1, k1a);
  Boost(r1b, q1, k1b);
  Boost(r0a, q0, k0a);
  Boost(r0b, q0, k0b);
  return fac;
}

double jacroooo(double s, double m1, double g1,
	double m0a, cvector4 k0a, double m0b, cvector4 k0b,
	double m1a, cvector4 k1a, double m1b, cvector4 k1b)
{
  double s1 = m1a*m1a + m1b*m1b + 2*Dot(k1a,k1b),
         s0 = m0a*m0a + m0b*m0b + 2*Dot(k0a,k0b),
         ssq = sqrt(s);
  return 2*s * jacResonance(s1, m1, g1, ssq)
	     * jacFlat(s0, ssq-sqrt(s1)) /
	  	( jacDec(s, s1, s0)
	  	* jacDec(s1, m1a*m1a, m1b*m1b)
	  	* jacDec(s0, m0a*m0a, m0b*m0b) );
}

  
double Maporooo(double *xran, double s, double m1, double g1,
	double m0a, vector4 k0a, double m1a, vector4 k1a,
	double m1b, vector4 k1b, double m1c, vector4 k1c)
{
  static double q1[4], q2[4], r1[4], r2[4], r2a[4], r2b[4];
  double fac, s1sq, s2sq, ssq = sqrt(s);

  fac = Resonance(xran[0], m1, g1, ssq-m0a, s1sq);
  fac *=Flat(xran[1], s1sq-m1a, s2sq);
  fac *= 1/(2*s) * SubDec(xran+2, ssq, s1sq, q1, m0a, k0a)
	  	 * SubDec(xran+4, s1sq, s2sq, r2, m1a, r1)
	  	 * SubDec(xran+6, s2sq, m1b, r2a, m1c, r2b);
  Boost(r1, q1, k1a);
  Boost(r2, q1, q2);
  Boost(r2a, q2, k1b);
  Boost(r2b, q2, k1c);
  return fac;
}

double jacorooo(double s, double m1, double g1,
	double m0a, cvector4 k0a, double m1a, cvector4 k1a,
	double m1b, cvector4 k1b, double m1c, cvector4 k1c)
{
  double ssq = sqrt(s);
  double s1 = s + m0a*m0a - 2* ssq * k0a[0],	    // s1 = ((p1+p2) - k0a)^2
         s2 = m1b*m1b + m1c*m1c + 2*Dot(k1b,k1c);
  return 2*s * jacResonance(s1, m1, g1, ssq-m0a)
	     * jacFlat(s2, sqrt(s1)-m1a) /
	  	( jacDec(s, s1, m0a*m0a)
	  	* jacDec(s1,s2, m1a*m1a)
	  	* jacDec(s2, m1b*m1b, m1c*m1c) );
}

  
double Maprroooo(double *xran, double s,
  	double m1, double g1, double m2, double g2,
	double m1a, vector4 k1a, double m1b, vector4 k1b,
	double m2a, vector4 k2a, double m2b, vector4 k2b)
{
  static double q1[4], q2[4], r1a[4], r1b[4], r2a[4], r2b[4];
  double fac, s1sq, s2sq, ssq = sqrt(s);
  
  if(ssq < PRTHR_CUTREL*(m1+m2))		// too far below threshold
    return Mapoooo(xran, s, m1a, k1a, m1b, k1b, m2a, k2a, m2b, k2b);

  fac = Resonance(xran[0], m1, g1, ssq, s1sq) *
        Resonance(xran[1], m2, g2, ssq, s2sq);
  fac *= 1/(2*s) * SubDec(xran+2, ssq, s1sq, q1, s2sq, q2)
	  	 * SubDec(xran+4, s1sq, m1a, r1a, m1b, r1b)
	  	 * SubDec(xran+6, s2sq, m2a, r2a, m2b, r2b);
  Boost(r1a, q1, k1a);
  Boost(r1b, q1, k1b);
  Boost(r2a, q2, k2a);
  Boost(r2b, q2, k2b);
  return fac;
}

double jacrroooo(double s, double m1, double g1, double m2, double g2,
	double m1a, cvector4 k1a, double m1b, cvector4 k1b,
	double m2a, cvector4 k2a, double m2b, cvector4 k2b)
{
  if(s < sqr(PRTHR_CUTREL*(m1+m2)))	// too far below threshold
    return jacoooo(s, m1a, k1a, m1b, k1b, m2a, k2a, m2b, k2b);

  double s1 = m1a*m1a + m1b*m1b + 2*Dot(k1a,k1b),
         s2 = m2a*m2a + m2b*m2b + 2*Dot(k2a,k2b),
         ssq = sqrt(s);
  return 2*s * jacResonance(s1, m1, g1, ssq)
	     * jacResonance(s2, m2, g2, ssq) /
	  	( jacDec(s, s1, s2)
	  	* jacDec(s1, m1a*m1a, m1b*m1b)
	  	* jacDec(s2, m2a*m2a, m2b*m2b) );
}

  
double Maprorooo(double *xran, double s,
  	double m1, double g1, double m2, double g2,
	double m0a, vector4 k0a, double m1a, vector4 k1a,
	double m2a, vector4 k2a, double m2b, vector4 k2b)
{
  static double q1[4], q2[4], r2a[4], r2b[4], r1[4], r2[4];
  double fac, s1sq, s2sq, ssq = sqrt(s);
  
//  Map13Decay(xran, s, m1, g1, m2, g2, m0a, k0a, m1a, k1a, m2a, k2a, m2b, k2b);
  fac = Resonance(xran[0], m1, g1, ssq-m0a, s1sq);
  fac *=Resonance(xran[1], m2, g2, s1sq-m1a, s2sq);
  fac *= 1/(2*s) * SubDec(xran+2, ssq, s1sq, q1, m0a, k0a)
	  	 * SubDec(xran+4, s1sq, s2sq, r2, m1a, r1)
	  	 * SubDec(xran+6, s2sq, m2a, r2a, m2b, r2b);
  Boost(r1, q1, k1a);
  Boost(r2, q1, q2);
  Boost(r2a, q2, k2a);
  Boost(r2b, q2, k2b);
  return fac;
}

double jacrorooo(double s, double m1, double g1, double m2, double g2,
	double m0a, cvector4 k0a, double m1a, cvector4 k1a,
	double m2a, cvector4 k2a, double m2b, cvector4 k2b)
{
  double ssq = sqrt(s);
  double s1 = s + m0a*m0a - 2* ssq * k0a[0],	    // s1 = ((p1+p2) - k0a)^2
         s2 = m2a*m2a + m2b*m2b + 2*Dot(k2a,k2b);
  return 2*s * jacResonance(s1, m1, g1, ssq-m0a)
	     * jacResonance(s2, m2, g2, sqrt(s1)-m1a) /
	  	( jacDec(s, s1, m0a*m0a)
	  	* jacDec(s1,s2, m1a*m1a)
	  	* jacDec(s2, m2a*m2a, m2b*m2b) );
}


double Mapooooo(double *xran, double s,
	double m1, vector4 k1, double m2, vector4 k2, double m3, vector4 k3,
	double m4, vector4 k4, double m5, vector4 k5)
{
  static double q1[4], q2[4], qq1[4], rq1[4],
	  	r1[4], r2[4], r3[4], r4[4], r5[4];
  double fac, s1sq, s2sq, s11sq, ssq = sqrt(s);

  fac = Flat(xran[0], ssq, s1sq);
  fac *= Flat(xran[1], ssq-s1sq, s2sq);
  fac *= Flat(xran[2], s1sq, s11sq);
  fac *= 1/(2*s) * SubDec(xran+3, ssq, s1sq, q1, s2sq, q2)
                 * SubDec(xran+5, s1sq, m1, r1, s11sq, rq1)
                 * SubDec(xran+7, s2sq, m4, r4, m5, r5)
                 * SubDec(xran+9,s11sq, m2, r2, m3, r3);
  Boost(r1, q1, k1);
  Boost(r4, q2, k4);
  Boost(r5, q2, k5);
  Boost(rq1, q1, qq1);
  Boost(r2, qq1, k2);
  Boost(r3, qq1, k3);
  return fac;
}

double jacooooo(double s,
	double m1, cvector4 k1, double m2, cvector4 k2, double m3, cvector4 k3,
	double m4, cvector4 k4, double m5, cvector4 k5)
{
  double s11 = m2*m2 + m3*m3 + 2*Dot(k2,k3),
         s1 = m1*m1 + 2*(Dot(k1,k2)+Dot(k1,k3)) + s11,
         s2 = m4*m4 + m5*m5 + 2*Dot(k4,k5),
         fac, ssq = sqrt(s);
  fac = jacFlat(s1, ssq)
      * jacFlat(s2, ssq-sqrt(s1))
      * jacFlat(s11, sqrt(s1));
  return 2*s * fac /
          ( jacDec(s, s1, s2)
          * jacDec(s1, m1*m1, s11)
          * jacDec(s2, m4*m4, m5*m5)
          * jacDec(s11, m2*m2, m3*m3) );
}


double Mapoooooo(double *xran, double s,
	double m1, vector4 k1, double m2, vector4 k2, double m3, vector4 k3,
	double m4, vector4 k4, double m5, vector4 k5, double m6, vector4 k6)
{
  static double q1[4], q2[4], qq1[4], qq2[4], rq1[4], rq2[4],
	  	r1[4], r2[4], r3[4], r4[4], r5[4], r6[4];
  double fac, s1sq, s2sq, s11sq, s21sq, ssq = sqrt(s);

  fac = Flat(xran[0], ssq, s1sq);
  fac *= Flat(xran[1], ssq-s1sq, s2sq);
  fac *= Flat(xran[2], s1sq, s11sq);
  fac *= Flat(xran[3], s2sq, s21sq);
  fac *= 1/(2*s) * SubDec(xran+4, ssq, s1sq, q1, s2sq, q2)
                 * SubDec(xran+6, s1sq, m1, r1, s11sq, rq1)
                 * SubDec(xran+8, s2sq, m2, r2, s21sq, rq2)
                 * SubDec(xran+10,s11sq,m3, r3, m4, r4)
                 * SubDec(xran+12,s21sq,m5, r5, m6, r6);
  Boost(r1, q1, k1);
  Boost(r2, q2, k2);
  Boost(rq1, q1, qq1);
  Boost(rq2, q2, qq2);
  Boost(r3, qq1, k3);
  Boost(r4, qq1, k4);
  Boost(r5, qq2, k5);
  Boost(r6, qq2, k6);
  return fac;
}

double jacoooooo(double s,
	double m1, cvector4 k1, double m2, cvector4 k2, double m3, cvector4 k3,
	double m4, cvector4 k4, double m5, cvector4 k5, double m6, cvector4 k6)
{
  double s11 = m3*m3 + m4*m4 + 2*Dot(k3,k4),
  	 s21 = m5*m5 + m6*m6 + 2*Dot(k5,k6),
         s1 = m1*m1 + 2*(Dot(k1,k3)+Dot(k1,k4)) + s11,
         s2 = m2*m2 + 2*(Dot(k2,k5)+Dot(k2,k6)) + s21,
         fac, ssq = sqrt(s);
  fac = jacFlat(s1, ssq)
      * jacFlat(s2, ssq-sqrt(s1))
      * jacFlat(s11, sqrt(s1))
      * jacFlat(s21, sqrt(s2));
  return 2*s * fac /
          ( jacDec(s, s1, s2)
          * jacDec(s1, m1*m1, s11)
          * jacDec(s2, m2*m2, s21)
          * jacDec(s11, m3*m3, m4*m4)
          * jacDec(s21, m5*m5, m6*m6) );
}


double Maproooooo(double *xran, double s, double m1, double g1,
        double m01, vector4 k01, double m02, vector4 k02,
        double m03, vector4 k03, double m04, vector4 k04,
	double m1a, vector4 k1a, double m1b, vector4 k1b)
{
  static double q1[4], q2[4], qq1[4], qq2[4], rq1[4], rq2[4],
	  	r1[4], r2[4], r3[4], r4[4], r5[4], r6[4];
  double fac, s1sq, s2sq, s11sq, s21sq, ssq = sqrt(s);

  fac = Flat(xran[0], ssq, s1sq);
  fac *= Flat(xran[1], ssq-s1sq, s2sq);
  fac *= Flat(xran[2], s1sq, s11sq);
  fac *= Resonance(xran[3], m1, g1, s2sq, s21sq);
  fac *= 1/(2*s) * SubDec(xran+4, ssq, s1sq, q1, s2sq, q2)
                 * SubDec(xran+6, s1sq, m01, r1, s11sq, rq1)
                 * SubDec(xran+8, s2sq, m02, r2, s21sq, rq2)
                 * SubDec(xran+10,s11sq,m03, r3, m04, r4)
                 * SubDec(xran+12,s21sq,m1a, r5, m1b, r6);
  Boost(r1, q1, k01);
  Boost(r2, q2, k02);
  Boost(rq1, q1, qq1);
  Boost(rq2, q2, qq2);
  Boost(r3, qq1, k03);
  Boost(r4, qq1, k04);
  Boost(r5, qq2, k1a);
  Boost(r6, qq2, k1b);
  return fac;
}

double jacroooooo(double s, double m1, double g1,
        double m01, cvector4 k01, double m02, cvector4 k02,
        double m03, cvector4 k03, double m04, cvector4 k04,
	double m1a, cvector4 k1a, double m1b, cvector4 k1b)
{
  double s11 = m03*m03 + m04*m04 + 2*Dot(k03,k04),
  	 s21 = m1a*m1a + m1b*m1b + 2*Dot(k1a,k1b),
         s1 = m01*m01 + 2*(Dot(k01,k03)+Dot(k01,k04)) + s11,
         s2 = m02*m02 + 2*(Dot(k02,k1a)+Dot(k02,k1b)) + s21,
         fac, ssq = sqrt(s);
  fac = jacFlat(s1, ssq)
      * jacFlat(s2, ssq-sqrt(s1))
      * jacFlat(s11, sqrt(s1))
      * jacResonance(s21, m1, g1, sqrt(s2));
  return 2*s * fac /
          ( jacDec(s, s1, s2)
          * jacDec(s1, m01*m01, s11)
          * jacDec(s2, m02*m02, s21)
          * jacDec(s11, m03*m03, m04*m04)
          * jacDec(s21, m1a*m1a, m1b*m1b) );
}

double Maprroooooo(double *xran, double s,
  	double m1, double g1, double m2, double g2,
        double m01, vector4 k01, double m02, vector4 k02,
        double m1a, vector4 k1a, double m1b, vector4 k1b,
	double m2a, vector4 k2a, double m2b, vector4 k2b)
{
  static double q1[4], q2[4], qq1[4], qq2[4], rq1[4], rq2[4],
	  	r1[4], r2[4], r3[4], r4[4], r5[4], r6[4];
  double fac, s1sq, s2sq, s11sq, s21sq, ssq = sqrt(s);

  fac = Flat(xran[0], ssq, s1sq);
  fac *= Flat(xran[1], ssq-s1sq, s2sq);
  fac *= Resonance(xran[2], m1, g1, s1sq, s11sq);
  fac *= Resonance(xran[3], m2, g2, s2sq, s21sq);
  fac *= 1/(2*s) * SubDec(xran+4, ssq, s1sq, q1, s2sq, q2)
                 * SubDec(xran+6, s1sq, m01, r1, s11sq, rq1)
                 * SubDec(xran+8, s2sq, m02, r2, s21sq, rq2)
                 * SubDec(xran+10,s11sq,m1a, r3, m1b, r4)
                 * SubDec(xran+12,s21sq,m2a, r5, m2b, r6);
  Boost(r1, q1, k01);
  Boost(r2, q2, k02);
  Boost(rq1, q1, qq1);
  Boost(rq2, q2, qq2);
  Boost(r3, qq1, k1a);
  Boost(r4, qq1, k1b);
  Boost(r5, qq2, k2a);
  Boost(r6, qq2, k2b);
  return fac;
}

double jacrroooooo(double s, double m1, double g1, double m2, double g2,
        double m01, cvector4 k01, double m02, cvector4 k02,
        double m1a, cvector4 k1a, double m1b, cvector4 k1b,
	double m2a, cvector4 k2a, double m2b, cvector4 k2b)
{
  double s11 = m1a*m1a + m1b*m1b + 2*Dot(k1a,k1b),
  	 s21 = m2a*m2a + m2b*m2b + 2*Dot(k2a,k2b),
         s1 = m01*m01 + 2*(Dot(k01,k1a)+Dot(k01,k1b)) + s11,
         s2 = m02*m02 + 2*(Dot(k02,k2a)+Dot(k02,k2b)) + s21,
         fac, ssq = sqrt(s);
  fac = jacFlat(s1, ssq)
      * jacFlat(s2, ssq-sqrt(s1))
      * jacResonance(s11, m1, g1, sqrt(s1))
      * jacResonance(s21, m2, g2, sqrt(s2));
  return 2*s * fac /
          ( jacDec(s, s1, s2)
          * jacDec(s1, m01*m01, s11)
          * jacDec(s2, m02*m02, s21)
          * jacDec(s11, m1a*m1a, m1b*m1b)
          * jacDec(s21, m2a*m2a, m2b*m2b) );
}
